FONT = "Roboto, Arial, sans-serif"

FONT_SIZE_TINY = 8
FONT_SIZE_SMALL = 12
FONT_SIZE_REGULAR = 14
FONT_SIZE_LARGE = 16
FONT_SIZE_EXTRA_LARGE = 20

FONT_WEIGHT_REGULAR = "normal"
FONT_WEIGHT_BOLD = "bold"
